/*######     Copyright (c) 1997-2014 Ufasoft  http://ufasoft.com  mailto:support@ufasoft.com,  Sergey Pavlov  mailto:dev@ufasoft.com #########################################################################################################
#                                                                                                                                                                                                                                            #
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation;  either version 3, or (at your option) any later version.          #
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.   #
# You should have received a copy of the GNU General Public License along with this program; If not, see <http://www.gnu.org/licenses/>                                                                                                      #
############################################################################################################################################################################################################################################*/

#include <el/inc/ext_config.h>
#include <el/libext/vc.h>

#if UCFG_WDM
#	include <el/libext.h>
#else
#	include <stdio.h>
#endif

#include <setjmp.h>
#include <stdint.h>

namespace Ext {}

using namespace Ext;

#if !UCFG_WDM && !UCFG_WCE
#	include <winternl.h>
#endif

#if UCFG_WIN32
#	include <windows.h>
#endif

#if UCFG_WCE && !UCFG_STDSTL
#	include <windows/ce/corecrtstorage.h>

extern "C" crtEHData* __crt_get_storage_EH();
#endif

#include "ext-os-api.h"

#if UCFG_LIB_DECLS && !defined(_WIN64) && UCFG_CRT!='U'
#	pragma comment(lib, "elrt")
#endif

#if UCFG_OS_IMPTLS										// to use our tlssup
#	if UCFG_PLATFORM_IX86
#		pragma comment(linker, "/INCLUDE:__tls_used")		
#	else
#		pragma comment(linker, "/INCLUDE:_tls_used")
#	endif
#endif


#if UCFG_WDM
#	include "../c-runtime/c-internal.h"


#endif


__BEGIN_DECLS

#if UCFG_WDM
	int _charmax = CHAR_MAX;
	__declspec(dllimport) void __stdcall RtlRaiseException(EXCEPTION_RECORD *er);
	__declspec(dllimport) void _stdcall RtlUnwind (void * TargetFrame, void *TargetIp, PEXCEPTION_RECORD ExceptionRecord, void * ReturnValue);

void * __cdecl API_memmove(void *dst, const void *src, size_t size) {
	RtlMoveMemory(dst, src, size);
	return dst;
}

#else

void * __cdecl API_memmove(void *dst, const void *src, size_t size) {
	return memmove(dst, src, size);
}

#endif

#if UCFG_WIN32_FULL
	typedef VOID (NTAPI *PFN_RtlRaiseException)(IN PEXCEPTION_RECORD);
	PFN_RtlRaiseException g_pfnRtlRaiseException = (PFN_RtlRaiseException)::GetProcAddress(::GetModuleHandleA("NTDLL.DLL"), "RtlRaiseException");
/*!!!R
	static struct InitPfnRtlRaiseException {
		InitPfnRtlRaiseException() {
			g_pfnRtlRaiseException = 
		}
	} s_InitPfnRtlRaiseException;	
	*/
	
	//static DlProcWrap<PFN_RtlRaiseException> g_pfnRtlRaiseException("NTDLL.DLL", "RtlRaiseException");
#endif

void _stdcall API_RtlRaiseException(EXCEPTION_RECORD *e) {
#if UCFG_WDM
	RtlRaiseException(e);
#else
	if (g_pfnRtlRaiseException)
		g_pfnRtlRaiseException(e);
	else
		RaiseException(e->ExceptionCode, e->ExceptionFlags, e->NumberParameters, e->ExceptionInformation);
#endif
}

void _stdcall API_RaiseException(DWORD dwExceptionCode, DWORD dwExceptionFlags, DWORD nNumberOfArguments, CONST ULONG_PTR *lpArguments) {
#if UCFG_WDM
	EXCEPTION_RECORD er = { (NTSTATUS)dwExceptionCode, dwExceptionFlags, 0, 0, nNumberOfArguments };
	memcpy(er.ExceptionInformation, lpArguments, nNumberOfArguments*sizeof(ULONG_PTR));
	RtlRaiseException(&er);
#else
	RaiseException(dwExceptionCode, dwExceptionFlags, nNumberOfArguments, lpArguments);
#endif
}

#if UCFG_WCE

HRESULT API_SafeArrayGetVartype(SAFEARRAY * psa, VARTYPE * pvt) {
	if (!psa || !pvt)
		return E_INVALIDARG;
	USHORT fFeatures = psa->fFeatures;
	if (!(fFeatures & (FADF_HAVEIID|FADF_HAVEVARTYPE)))
		return E_INVALIDARG;
	if (fFeatures & FADF_RECORD)
		*pvt = VT_RECORD;
	else if (fFeatures & FADF_HAVEIID)
		*pvt = VARTYPE((fFeatures & FADF_DISPATCH) ? VT_DISPATCH : VT_UNKNOWN);
	else
		*pvt = VARTYPE(((DWORD*)psa)[-1]);
	return 0;
}

void __stdcall xxx_RaiseException(DWORD dwExceptionCode, DWORD dwExceptionFlags, DWORD nNumberOfArguments, CONST ULONG_PTR *lpArguments) {
	RaiseException(dwExceptionCode, dwExceptionFlags, nNumberOfArguments, lpArguments);
}

BOOL __stdcall KernelLibIoControl(HANDLE hLib, DWORD dwIoControlCode, LPVOID lpInBuf, DWORD nInBufSize, LPVOID lpOutBuf, DWORD nOutBufSize, LPDWORD lpBytesReturned);
BOOL __stdcall xxx_KernelLibIoControl(HANDLE hLib, DWORD dwIoControlCode, LPVOID lpInBuf, DWORD nInBufSize, LPVOID lpOutBuf, DWORD nOutBufSize, LPDWORD lpBytesReturned) {
	return KernelLibIoControl(hLib, dwIoControlCode, lpInBuf, nInBufSize, lpOutBuf, nOutBufSize, lpBytesReturned);
}
#endif // UCFG_WCE

#if defined(_M_IX86) && defined(_MSC_VER) //!!!!UCFG_STDSTL

void _stdcall API_RtlUnwind (void * TargetFrame, void *TargetIp, PEXCEPTION_RECORD ExceptionRecord, void * ReturnValue) {
#if UCFG_WDM
	RtlUnwind(TargetFrame, TargetIp, ExceptionRecord, ReturnValue);
#else
	IMP_RtlUnwind(TargetFrame, TargetIp, ExceptionRecord, ReturnValue);
#endif
}

#endif

#ifdef _M_X64

#ifndef UNWIND_HISTORY_TABLE_SIZE
	typedef void *PUNWIND_HISTORY_TABLE;		// defined in new SDK

	extern "C" NTSYSAPI PVOID NTAPI RtlPcToFileHeader(PVOID PcValue, PVOID *BaseOfImage);
	extern "C" VOID NTAPI  RtlUnwindEx (void *TargetFrame, PVOID TargetIp, PEXCEPTION_RECORD ExceptionRecord, PVOID ReturnValue, PCONTEXT OriginalContext, void *HistoryTable);
	extern "C" PVOID WINAPI RtlLookupFunctionEntry(ULONGLONG ControlPC, PULONGLONG ImageBase, void *TargetGp);
#endif

#if !UCFG_EH_OS_UNWIND
	
	extern "C" VOID NTAPI My_RtlUnwindEx(
    _In_opt_ PVOID TargetFrame,
    _In_opt_ PVOID TargetIp,
    _In_opt_ PEXCEPTION_RECORD ExceptionRecord,
    _In_ PVOID ReturnValue,
    _In_ PCONTEXT ContextRecord,
    _In_opt_ PUNWIND_HISTORY_TABLE HistoryTable
    );
#	define RtlUnwindEx My_RtlUnwindEx
#endif

void * _stdcall API_RtlPcToFileHeader(const void *PcValue, void **pBaseOfImage) {
	return RtlPcToFileHeader((void*)PcValue, pBaseOfImage);
}

void _stdcall API_RtlUnwindEx(void *TargetFrame, PVOID TargetIp, PEXCEPTION_RECORD ExceptionRecord, PVOID ReturnValue, PCONTEXT OriginalContext, void *HistoryTable) {
	RtlUnwindEx(TargetFrame, TargetIp, ExceptionRecord, ReturnValue, OriginalContext, (PUNWIND_HISTORY_TABLE)HistoryTable);
}

void * _stdcall API_RtlLookupFunctionEntry(DWORDLONG ControlPC, DWORDLONG *ImageBase, DWORDLONG *TargetGp) {
	return RtlLookupFunctionEntry(ControlPC, ImageBase, (PUNWIND_HISTORY_TABLE)TargetGp);
}

#endif

#if !UCFG_WCE

void * _stdcall API_SetUnhandledExceptionFilter(void *lpTopLevelExceptionFilter) {
#if UCFG_WDM
	Throw(E_FAIL);
#else
	return (LPTOP_LEVEL_EXCEPTION_FILTER )SetUnhandledExceptionFilter((LPTOP_LEVEL_EXCEPTION_FILTER )lpTopLevelExceptionFilter);
#endif
}
#endif

BOOL _stdcall API_IsBadReadPtr(CONST VOID *lp, UINT_PTR ucb) {
#if UCFG_WDM
	return !MmIsAddressValid((void*)lp);
#else
	return IsBadReadPtr(lp, ucb);
#endif
}

BOOL _stdcall API_IsBadWritePtr(VOID *lp, UINT_PTR ucb) {
#if UCFG_WDM
	return !MmIsAddressValid(lp);
#else
	return IsBadWritePtr(lp, ucb);
#endif
}

BOOL _stdcall API_IsBadCodePtr(void *lp) {
#if UCFG_WDM
	return !MmIsAddressValid(lp);
#else
	return IsBadCodePtr((FARPROC)lp);
#endif
}

BOOLEAN _stdcall API_RtlpGetStackLimits(void **stackLimit, void **stackBase) {
#if UCFG_WCE
	const TEB& teb = *(TEB*)(*(byte**)PUserKData - sizeof(TEB));
	*stackLimit = teb.StackBound;
	*stackBase = (byte*)teb.StackBase +teb.ProcStackSize;
#elif defined(WIN32) //!!! was _WIN32
	const NT_TIB& tib = *(NT_TIB*)NtCurrentTeb();
	*stackBase = tib.StackBase;
	*stackLimit = tib.StackLimit;
#else
	Throw(E_NOTIMPL);
#endif;
	const void *sp = &sp;
	return sp >= *stackLimit && sp < *stackBase;
}


#if UCFG_WDM
	PFN_printf g_API_printf = &DbgPrint;
#else
	PFN_printf g_API_printf = (PFN_printf)&printf;
#endif

void _cdecl API_terminate() {
#if UCFG_WDM
	KeBugCheck(123); //!!!
#elif UCFG_WCE
	ExitProcess(3);
#else
	::terminate();
#endif
}

void _cdecl API_unexpected() {
#if UCFG_WDM
	KeBugCheck(124); //!!!
#elif UCFG_WCE
	ExitProcess(3);
#else
	::unexpected();
#endif
}

bool _cdecl API_uncaught_exception() {
#if UCFG_STDSTL
	abort(); //!!!
	return false;
#else
	return Ext::ExcRefs::ProcessingThrowRef();
#endif
}


#ifdef _M_IX86

__declspec(dllexport) __declspec(naked) void __cdecl My__ehvec_copy_ctor() {
	_asm  {
		jmp __ehvec_copy_ctor
	}
}

#endif

__END_DECLS



namespace Ext {

#if !UCFG_WDM

PFN_RtlDispatchException __stdcall GetProcAddress_RtlDispatchException() {
#	if UCFG_PLATFORM_X64 && !UCFG_WDM
	const byte sig[8] = { 0x48, 0x8D, 0x55, 0x20, 0x48, 0x8B, 0xCB, 0xE8 };
	if (byte *pRtlRaiseException = (byte*)GetProcAddress(GetModuleHandleW(L"NTDLL.dll"), "RtlRaiseException")) {
		for (int i=100; i<1000; ++i) {
			if (!memcmp(pRtlRaiseException+i, sig, sizeof sig)) {
				byte *addr = pRtlRaiseException + i + sizeof(sig);
				return PFN_RtlDispatchException(addr + 4 + *(int32_t*)addr);
			}
		}
	}
#	endif
	return 0;
}

#endif //!UCFG_WDM

#if UCFG_WCE && !UCFG_STDSTL
ExcRefs ExcRefs::GetForCurThread() {
	crtEHData& ehd = *__crt_get_storage_EH();
	return ExcRefs((EHExceptionRecord*&)ehd.CurrentExcept, ehd.ExceptContext);
}

int& ExcRefs::ProcessingThrowRef() {
	return __crt_get_storage_EH()->_ProcessingThrow;
}

#endif

#if UCFG_EXT_C
ExcRefs ExcRefs::GetForCurThread() {
	C_ThreadStorage& ts = C_ThreadStorage::GetForCurThread();
	return ExcRefs(ts.ExceptionRecord, (void*&)ts.Context, ts.FrameInfoChain, ts.m_curexcspec, ts.CxxReThrow, ts._translator
#	ifdef _M_X64
	, ts._curexception
	, ts._curcontext
	, ts._imageBase
	, ts._throwImageBase
	, ts._foreignException
#	endif	
	);
}

int& ExcRefs::ProcessingThrowRef() {
	return C_ThreadStorage::GetForCurThread()._ProcessingThrow;
}


#endif


#ifdef _M_X64
extern "C" {
DECLSPEC_NORETURN VOID NTAPI RtlRaiseStatus(IN NTSTATUS Status);
void __stdcall RtlRaiseException(EXCEPTION_RECORD *er);
NTSYSAPI NTSTATUS NTAPI ZwRaiseException(IN PEXCEPTION_RECORD ExceptionRecord, IN PCONTEXT Context, IN BOOLEAN SearchFrames);
NTSYSAPI NTSTATUS NTAPI ZwContinue(IN PCONTEXT Context, IN BOOLEAN TestAlert);

static PFN_RtlDispatchException s_pfnRtlDispatchException;
static volatile bool s_bTriedGetProcRtlDispatchException;

#if UCFG_WDM

VOID NTAPI API_TryMyRtlRaiseException(IN PEXCEPTION_RECORD ExceptionRecord) {
}

#else


VOID NTAPI API_TryMyRtlRaiseException(IN PEXCEPTION_RECORD ExceptionRecord) {
	if (!s_bTriedGetProcRtlDispatchException) {
		s_pfnRtlDispatchException = GetProcAddress_RtlDispatchException();
		s_bTriedGetProcRtlDispatchException = true;
	}
	if (!s_pfnRtlDispatchException)
		return;
    NTSTATUS Status = STATUS_INVALID_DISPOSITION;
    CONTEXT Context;
    ULONG64 ImageBase;
    RtlCaptureContext(&Context);
    if (PRUNTIME_FUNCTION  FunctionEntry = RtlLookupFunctionEntry(Context.Rip, &ImageBase, NULL)) {
	    PVOID HandlerData;
		ULONG64 EstablisherFrame;
		RtlVirtualUnwind(UNW_FLAG_NHANDLER, ImageBase, Context.Rip, FunctionEntry, &Context, &HandlerData, &EstablisherFrame, NULL);
		ExceptionRecord->ExceptionAddress = (PVOID)Context.Rip;
		Context.ContextFlags = CONTEXT_FULL;
		Status = s_pfnRtlDispatchException(ExceptionRecord, &Context)
			? ZwContinue(&Context, FALSE)
			: ZwRaiseException(ExceptionRecord, &Context, FALSE);
    }
    RtlRaiseStatus(Status);
	API_terminate();		//!!! not reachable
}

#endif // UCFG_WDM

} // "C"

#endif // _M_X64



} // Ext::

#ifdef _CRTBLD
extern "C" void* __cdecl API_malloc(size_t size) {
	return malloc(size);
}
#endif


